from meta import *
from modules import defaults, cli, gui as gui_handler
from modules.clicklog import Overview, info, warn
from typing import IO
import datetime
import logging
import click
import os


@click.group()
@click.version_option(version=VERSION, prog_name=TOOL_NAME)
def beads() -> None:
    """
    Simple Code generation tool parsing fine state machines and transpiling them into code snippets.

    Beads provides a commandline interface and a graphical user interface (GUI) via their
    respective commands.

    Provide or draw a fine state machine (fsm) and code corresponding to the fsm will be generated. An overview of all
    supported programming languages can be printed using the 'languages' command on the commandline,
    or viewed under 'Settings' in the GUI.

    The list of all available command and options to use this tool can be found online:
    https://gitlab.beuth-hochschule.de/s40242/stategen/blob/master/README.md

    and on the commandline upon providing the '--help' option like so: 'beads --help' or 'beads [any command] --help'
    for each sub-command respectively.

    The GUI also provides a readme and help section here: 'Menu' > 'Tool Info'
    """

    pass


@beads.command()
@click.option('-b', '--background', 'no_window', flag_value=True, help=help_no_window)
@click.option('-p', '--port', type=click.INT, default=None, help=help_port)
@click.option('-v', '--verbose', 'is_verbose', flag_value=True, help=help_verbose)
@click.option('--file', type=click.File(mode='r'), default=None, help=help_file_gui)
def gui(no_window: bool, port: int,  is_verbose: bool, file: IO) -> None:
    """
    Open a graphical user interface (GUI).

    The GUI allows drawing of fine state machines and provides means to set/change user specific default settings
    that are considered in commandline usage as well.
    """

    opts = dict()
    opts[NO_WINDOW] = no_window or defaults.get(NO_WINDOW, False)
    opts[PORT] = port or defaults.get(PORT, 8000)

    set_up_logger(is_verbose)

    gui_handler.start(file, opts)


@beads.command()
@click.option('-l', '--language', type=click.Choice(supported_languages), default=None, help=help_languages)
@click.option('-nv', '--no-validation', flag_value=True, help=help_no_validation)
@click.option('-o', '--out', 'file_name', type=str, default=None, help=help_compile_name)
@click.option('-bd', '--base-directory', type=click.Path(exists=True, file_okay=False), default=None, help=help_out_dir)
@click.option('-re', '--replace-existing', 'replace', flag_value=True, help=help_replace_existing)
@click.option('-v', '--verbose', 'is_verbose', flag_value=True, help=help_verbose)
@click.argument('files', type=click.File(mode='r'), nargs=-1, required=True)
def parse(files: list, language: str, no_validation: bool, is_verbose: bool,
          base_directory: str, file_name: str, replace: bool) -> None:
    """
    Parse a provided FILE into code.

    Parse the provided FILE containing a textual representation of a fine state machine
    and generate a code snippet representing this machines internal logic.
    """

    set_up_logger(is_verbose)

    opts = dict()
    opts[SKIP_VALIDATION] = no_validation or defaults.get(SKIP_VALIDATION, False)
    opts[REPLACE] = replace or defaults.get(REPLACE, False)
    opts[OUT_DIR] = base_directory or defaults.get(OUT_DIR, os.getcwd())

    language = language or defaults.get(LANGUAGE, None)

    if len(files) == 1:
        file = files[0]

        opts[FILENAME] = file_name or default_file_name(file)

        cli.parse(file, language, opts)

    else:
        for file in files:
            opts[FILENAME] = default_file_name(file)

            cli.parse(file, language, opts)


@beads.command()
@click.argument('search_text', type=str, default=None, required=False)
def languages(search_text: str) -> None:
    """
    Print an overview of all supported programming languages.

    Languages can be filter by providing a SEARCH_TEXT to look for.

    :param search_text: String that is contained in language name or None
    """

    end: str = f' with "{search_text}"' if search_text is not None else ''
    title: str = f'Supported programming languages{end}'

    overview: Overview = Overview(title)
    overview.print(f'*   {language}' for language in filter_languages(search_text))


@beads.command()
@click.option('-s', '--set-default', 'options_to_set', type=(str, str), multiple=True, help=help_set_option)
@click.option('-u', '--unset-default', 'options_to_unset', type=str, multiple=True, help=help_unset_option)
@click.option('--unset-all', flag_value=True, help=help_unset_all)
@click.option('--show', flag_value=True, help=help_show_options)
def options(options_to_set: list, options_to_unset: list, unset_all: bool, show: bool) -> None:
    """
    Print an overview of all currently set default runtime options.

    Allows to view, set and unset options that are used by default.
    Provide --show to print a list of all available options and their value types.

    Use -s/--set as:   --set language python

    Use -u/--unset as:    --unset language

    Setting and Un-setting of options can be provided multiple times for one call.
    The options take the following precedence:

        1. If unset-all is given, set and unset will be ignored!

        2. Setting is done before un-setting!

        3. --show will be executed independently
    """

    if show:
        overview: Overview = Overview('Available default runtime options')
        overview.print(f'  {key}: {value.__name__}' for key, value in defaults.keys_values.items())

    if unset_all:
        if click.confirm('Delete all currently set default options?'):
            defaults.unset_all()
            info('Deleted all runtime default options.')
        else:
            info('Aborted!')

    elif len(options_to_set) > 0 or len(options_to_unset) > 0:
        for option, value in options_to_set:
            if defaults.set_default(option, value):
                info(f'Set --{option} to {value}!')
            else:
                warn(f'Could not set --{option} to {value}')

        for option in options_to_unset:
            if defaults.unset_default(option):
                info(f'Unset option --{option}!')
            else:
                warn(f'Could not unset --{option}')

        defaults.save()

    else:
        if len(defaults.current_defaults) == 0:
            info('There are no default runtime options set!')

        else:
            overview: Overview = Overview('Current default runtime options')
            overview.print(f'  --{key} {value}' for key, value in defaults.current_defaults.items())


def set_up_logger(verbose: bool) -> None:
    """
    Set up logging for program execution.

    Provide 'verbose' as True to set 'DEBUG' logging level.

    :param verbose: Debug flag
    """

    is_verbose = verbose or defaults.get(VERBOSE, False)

    root_logger = logging.getLogger()
    if root_logger.handlers:
        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)

    log_level = logging.DEBUG if is_verbose else logging.WARNING
    logging.basicConfig(level=log_level)
    logging.debug(f'MAIN: Set up logger with level={log_level}')


def default_file_name(file: IO) -> str:
    """
    Get the default name for an input file

    :param file: IO user input
    :return: default name for file
    """

    default_name = file.name.rpartition('.')[0]

    if default_name == '':
        default_name = f'generated_code_{str(datetime.datetime.now())}'

    return default_name


def filter_languages(search_text: str) -> iter:
    """
    Filter all supported languages by given SEARCH_TEXT.

    :param search_text: String to look for in supported languages
    :return: ALl languages that contain the search_text.
    """

    for language in sorted(supported_languages):
        if search_text is None or search_text in language:
            yield language


if __name__ == '__main__':
    exit(beads())



from meta import *
from modules import reader, writer
from distutils.util import strtobool
import logging


def set_default(option: str, value: str) -> bool:
    """
    Set a default runtime OPTION to the provided VALUE.

    Automatically handles int and bool values if needed.

    :param option: The OPTION to set the value for
    :param value: Value as string to set the OPTION to
    :return: True if option was successfully set, otherwise False
    """

    success: bool = False

    if option in keys_values.keys():
        new_value = value_for_key(option, value)

        if new_value is not None:
            current_defaults[option] = new_value
            logging.debug(f'DEFAULTS: Set {option} to {new_value}.')
            success = True

    return success


def unset_default(option: str) -> bool:
    """
    Un-set a default runtime OPTION.

    :param option: The OPTION to unset
    :return: True if option was successfully unset, otherwise False
    """

    success: bool = False

    if option in keys_values.keys():
        try:
            del current_defaults[option]
            logging.debug(f'DEFAULTS: Unset {option}')
            success = True

        except KeyError:
            pass

    return success


def unset_all() -> None:
    """
    Unset all currently saved default runtime options.

    Method will save the empty default-options!
    """

    writer.save_resource(OPTION_SUB_DIR, DEFAULT_OPTIONS, {})


def save() -> None:
    """
    Persist the currently set default runtime options.
    """

    writer.save_resource(OPTION_SUB_DIR, DEFAULT_OPTIONS, current_defaults)


def get(key: str, value_on_none: str or int or bool) -> str or int or bool:
    """
    Get the default value for the provided runtime option as KEY.
    If no such value is present the VALUE_ON_NONE will be returned.

    :param key: Runtime Option key to get value for
    :param value_on_none: Value for Option (KEY)
    :return: Set default value or VALUE_ON_NONE if none found
    """

    default_value = value_on_none

    if key in current_defaults:
        default_value = current_defaults[key]

    return default_value


def load_defaults() -> dict:
    """
    Load the saved defaults for runtime options.

    :return: Dictionary of defaults or empty if none could be found.
    """

    defaults: dict = {}

    for key, value in reader.load_resource(OPTION_SUB_DIR, DEFAULT_OPTIONS).items():
        if key in keys_values:
            value_to_set = value_for_key(key, value)

            if value_to_set is not None:
                defaults[key] = value_to_set

    return defaults


def value_for_key(key: str, value) -> str or int or bool or None:
    """
    Parse the provided value into the required form the the given key.

    Will return None if any problems occur.

    :param key: Key string for default options.
    :param value: Value to parse
    :return: Parsed value or None
    """
    new_value = None

    try:
        option_type = keys_values[key]

        if option_type is int:
            new_value = int(value)

        elif option_type is bool:
            if type(value) is str:
                value = strtobool(value)

            new_value = bool(value)

        elif option_type is str:
            new_value = str(value)

    except ValueError:
        logging.debug(f'DEFAULTS: Could not parse "{value}" into required form for key "{key}"!')

    return new_value


keys_values: dict = {
    LANGUAGE: str,
    NO_WINDOW: bool,
    OUT_DIR: str,
    PORT: int,
    REPLACE: bool,
    SKIP_VALIDATION: bool,
    VERBOSE: bool,
}

keys_help: dict = {
    LANGUAGE: help_languages,
    NO_WINDOW: f'{help_no_window} Requires new start to take effect.',
    OUT_DIR: help_out_dir,
    PORT: f'{help_port} Requires new start to take effect.',
    REPLACE: help_replace_existing,
    SKIP_VALIDATION: help_no_validation,
    VERBOSE: f'{help_verbose} Requires new start to take effect.'
}

current_defaults: dict = load_defaults()



import click


def info(text: str) -> None:
    """
    Convenience click.echo call with '[INFO]: ' prefix

    :param text: String to log on terminal
    """
    click.echo(f'[INFO]: {text}')


def warn(text: str) -> None:
    """
    Convenience click.secho with fg=yellow.

    Text will be prefixed with '[WARN]: ' and printed in yellow.

    :param text: String to log on terminal
    """
    click.secho(f'[WARN]: {text}', fg='yellow')


def error(text: str) -> None:
    """
    Convenience click.secho with fg=red.

    Text will be prefixed with '[ERROR]: ' and printed in red.

    :param text: String to log on terminal
    """
    click.secho(f'[ERROR]: {text}', fg='red')


class Overview:
    """
    Overview to print on CommandLine.

    Provide the title of the overview on initialization.
    """

    def __init__(self, title: str):
        self.title = title

    def print(self, info) -> None:
        """
        Print the Overview on the CommandLine with it title and the given INFO in its body.

        :param info: Generator for strings to print as overviews body
        """
        click.echo(f'[OVERVIEW]: {self.title}:')
        click.echo('|')
        for item in info:
            click.echo(f'| {item}')
        click.echo('|')



from meta import *
from modules.errors import ParsingError
from typing import IO
import markdown as md
import logging
import json
import os


def parse_file(file: IO) -> dict:
    """
    Parse the provided file into a dictionary.

    Method to parse input files into an internally usable version (dict).
    All supported file formats can be found here: TODO
    To parse string / text use 'parse_text' instead!

    :param file: IO to parse
    :return: Dictionary containing the file contents
    """

    if file.name.endswith('.json'):
        logging.debug(f'READER: Identified {file.name} as json: Calling json.load')
        content = json.load(file)

    elif file.name.endswith('.bead'):
        logging.debug(f'READER: Identified {file.name} as beads-file: parsing from bds to json')
        content = parse_beads(file)

    else:
        raise ParsingError('Unsupported file-type!')

    logging.debug(f'READER: Parsed file: {file.name} to :\n{content}')

    return content


def parse_text(text: str) -> dict:
    """
    Parse the provided text / string into a dictionary.

    Method to parse input text / str into an internally usable version (dict).
    All supported text formats can be found here: TODO
    To parse files use 'parse_file' instead!

    :param text: Text / string to parse
    :return: Dictionary containing the string contents
    """

    try:
        return json.loads(text)

    except json.JSONDecodeError:
        raise ParsingError


def parse_beads(file: IO) -> dict:
    """
    Parse the provided Beads file into a dictionary.

    Method to parse input file into an internally usable version (dict).

    :param file: IO to parse
    :return: Dictionary containing the string contents
    """

    graph_name: str = 'graph'
    nodes: set = set()
    initial_node: str or None = None
    transitions: list = []

    first_line = file.readline()
    file.seek(0)

    if first_line.strip().startswith('#!'):
        config = read_config(first_line)

        if START in config.keys():
            initial_node = config[START]

        if NAME in config.keys():
            graph_name = config[NAME]

    if initial_node is not None:
        nodes.add(initial_node)

    for line in strip_comments(file):
        [_from, transition, _to] = decode_line(line)
        transition = {LABEL: transition, FROM: _from, TO: _to}

        transitions.append(transition)
        nodes.add(_from)
        nodes.add(_to)

    content = {
        NAME: graph_name,
        NODES: wrap_node_names(nodes, initial_node),
        TRANSITIONS: transitions
    }

    return content


def decode_line(transition: str) -> tuple:
    """
    Decode a line of a .bead file

    :param transition: str - line to validate
    :return: list of names for LABEL, FROM and TO
    """

    try:
        trans_strings = transition.strip().split(':')

        # assert transition string has valid structure
        assert len(trans_strings) == 3,\
            f'Invalid transition format\n expected: state_from:transition:state_to\n got: {transition}'

        trans_strings = (trans_strings[0].strip(), trans_strings[1].strip(), trans_strings[2].strip())

        # assert names are valid
        for string in trans_strings:
            for char in string:
                assert char.isalnum() or char == '_',\
                    f'Invalid sign "{char}" in "{string}":\nUse only alphanumeric signs (a - Z, 0 - 9) and "_"'

    except AssertionError as ae:
        logging.debug(f'READER: Decoding of {transition} failed!')
        raise ParsingError(ae)

    return trans_strings


def wrap_node_names(nodes: set, initial_node: str or None) -> list:
    """
    Wrap all node names into required dicts.

    :param initial_node: Optional node defining starting point
    :param nodes: Set of node names
    """

    nodes_list: list = []

    for node_name in sorted(nodes):
        node = {ID: node_name}

        if initial_node and node_name == initial_node:
            node[START] = True

        nodes_list.append(node)

    return nodes_list


def load_resource(subdir: str, name: str) -> dict:
    """
    Load a resource and provide the content as a DICT.

    This method will load a resource json file from the provided SUBDIR (directory)

    :param subdir: Subdirectory the resource is located in
    :param name: Filename of the resource
    :return: A parsed dictionary of the specified resource
    """

    resource_path = os.path.join(RESOURCES_PATH, subdir, name)
    logging.debug(f'READER: Constructed resource_path: {resource_path}')

    res: dict = {}
    if name.endswith(".json") or name.endswith(".template"):
        with open(resource_path, encoding=STANDARD_ENCODING) as resource:
            content: str = ''.join(strip_comments(resource))
            logging.debug(f'READER: Read content: {content}')

            res = json.loads(content)

    elif name.endswith(".md"):

        with open(resource_path, encoding=STANDARD_ENCODING) as resource:
            content: str = md.markdown(resource.read(), extensions=['fenced_code'])
            logging.debug(f'READER: Read content: {content}')

        res[CONTENT] = content

    logging.debug(f'READER: Loaded resource as: {res}')

    return res


def read_config(line: str) -> dict:
    """
    Read the config line of a file.

    Config lines start with #! and declare configuration item like name:value.
    Items need to be separated by whitespace.

    :param line: Config line as str
    :return: parsed config as dict
    """

    config: dict = {}

    config_list: list = line.split(' ')

    for element in config_list:
        if ':' in element:
            item = element.split(':')
            config.update({item[0]: item[1].replace('\n', '')})

    return config


def strip_comments(file: IO) -> str:
    """
    Clear a resource file of comments.

    This method will return a generator object yielding only lines that are striped from comments.
    Comments are cleared like so:
        # Comments outside will be cleared
        {
            # Single line comments inside will be cleared
            "key":"value" # Trailing comments will be cleared

            "key": # This will result in an error "value"
            "key#": This will result in an error "value"
            "key": "val '# This will also result in an error' ue"
        }

    :param file: Resource file to strip comments from
    :return: Generator for str
    """

    for line in file.readlines():
        line = line.partition('#')[0].rstrip()
        if line:
            yield line



from modules.errors import ValidationError, UnsupportedLanguageError
from modules.generator import CodeGenerator
from modules import templates, validator
import logging


def process(fsm: dict, language: str, skip_validation: bool = False) -> str:
    """
    Process the FSM in the given LANGUAGE.

    Set SKIP_VALIDATION argument to True if validation of internal fsm logic shall be skipped.
    Process pipeline will first validate the internal fsm logic and then generate code based on
    a template for the given language.

    :param fsm: Fine State Machine parsed to Dictionary
    :param language: String for programming language to use as base for code generation
    :param skip_validation: Flag to skip validation of fsm logic
    :return: Code string representing the provided fine state machine
    """

    logging.debug('LOGIC: Starting to parse Fine State Machine')

    try:
        if skip_validation:
            logging.debug('LOGIC: Skip validation of state machine logic')

        else:
            logging.debug('LOGIC: Start validation of state machine')
            validator.inspect(fsm)
            logging.debug('LOGIC: Validation of state machine logic successful')

        template: dict = templates.get_template(language)
        logging.debug(f'LOGIC: For language: "{language}" got template:\n{template}')

        code = CodeGenerator(template).build(fsm)
        logging.debug(f'LOGIC: Code generation ran successful.')

        logging.debug('LOGIC: Finished to parse Fine State Machine')
        return code

    except ValidationError as ve:
        raise ve
    except UnsupportedLanguageError as ule:
        raise ule



#!/usr/bin/env python3
import unittest
import modules.validator as val
from modules.errors import ValidationError


class ValidationTest(unittest.TestCase):
    """
    Class to test the internal validation of state machine logic.

    Rudimentary test coverage for common cases.
    """

    def test_a_valid_machine(self):
        """
        Test that a valid machine does not raise any errors on inspection.
        """

        correct_machine: dict = {
            "name": "correct machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {"id": "B"},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        val.inspect(correct_machine)

        self.assertTrue(True)

    def test_invalid_machine_with_missing_name(self):
        """
        Test that a missing name key raises a Validation error.
        """

        invalid_machine: dict = {
            # no name
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {"id": "B"},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        self.assertRaises(ValidationError, lambda: val.inspect(invalid_machine))

    def test_invalid_machine_with_missing_nodes(self):
        """
        Test that a missing nodes key raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            # no nodes
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        self.assertRaises(ValidationError, lambda: val.inspect(invalid_machine))

    def test_invalid_machine_with_missing_transitions(self):
        """
        Test that a missing transitions key raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {"id": "B"},
            ]
            # no transitions
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_empty_nodes(self):
        """
        Test that no declared node raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                # empty nodes
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_empty_transitions(self):
        """
        Test that no declared transition raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {"id": "B"},
            ],
            "transitions": [
                # empty transitions
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_node_missing_id(self):
        """
        Test that a node with missing id key raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {
                    # empty node
                },
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_duplicate_nodes(self):
        """
        Test that a duplicate node declaration raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {
                    # duplicate id
                    "id": "C"
                },
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_transitions_missing_keys(self):
        """
        Test that a duplicate node declaration raises a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                {"id": "B"},
            ],
            "transitions": [
                # missing 'label'
                {"from": "A", "to": "B"},
                # missing 'to'
                {"from": "A", "label": "Transition2"},
                # missing 'from'
                {"label": "Transition1", "to": "C"},
                # missing all required keys
                {},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_multiple_start_nodes(self):
        """
        Test that multiple start nodes raise a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "C"},
                {"id": "A", "start": True},
                # second start node
                {"id": "B", "start": True},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "C"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_unreachable_nodes(self):
        """
        Test that unreachable nodes raise a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "A", "start": True},
                {"id": "B"},
                # unreachable nodes C and D
                {"id": "C"},
                {"id": "D"}
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "B", "label": "Transition1", "to": "B"},
                {"from": "A", "label": "Transition2", "to": "A"},
                # only transition references of C and D
                {"from": "C", "label": "Transition3", "to": "D"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_duplicate_transitions(self):
        """
        Test that nodes that are referenced by the same transition multiple times as from node raise a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "A", "start": True},
                {"id": "B"},
                {"id": "C"},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "B", "label": "Transition1", "to": "C"},
                # second time Transition1 references A as from node
                {"from": "A", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_undeclared_node_reference(self):
        """
        Test that a reference to an undeclared node raise a Validation error.
        """

        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "A", "start": True},
                {"id": "B"},
                {"id": "C"},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
                # transitions that references the undeclared node D
                {"from": "A", "label": "Transition4", "to": "D"},
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)

    def test_invalid_machine_with_misspelled_node_reference(self):
        """
        Test that misspelled references to nodes raise a Validation error.
        """
        invalid_machine: dict = {
            "name": "invalid machine",
            "nodes": [
                {"id": "A", "start": True},
                {"id": "B"},
                {"id": "C"},
            ],
            "transitions": [
                {"from": "A", "label": "Transition1", "to": "B"},
                {"from": "B", "label": "Transition1", "to": "C"},
                {"from": "C", "label": "Transition3", "to": "A"},
                # transitions that references the undeclared node 'b' instead of 'B'
                {"from": "b", "label": "Transition2", "to": "A"}
            ]
        }

        with self.assertRaises(ValidationError):
            val.inspect(invalid_machine)



from meta import SKIP_VALIDATION, FILENAME, REPLACE, OUT_DIR, NAME
from modules.logic import process
from modules.templates import append_language_association
from modules.errors import ParsingError, InternalException, ExistingFileError, ValidationError, UnsupportedLanguageError
from modules.clicklog import info, warn, error
from modules import writer, reader
from typing import IO
import logging
import os


def parse(file: IO, language: str, opts: dict) -> None:
    """
    Parse the given file in the provided language and generate code.

    :param file: File to parse
    :param language: Programming language to use for code generation
    :param opts: Dictionary of available options
    """

    logging.debug(f'CLI: parse called with file: {file.name} and language: {language}')
    logging.debug(f'CLI: provided options: {opts}')

    try:
        fsm: dict = reader.parse_file(file)
        logging.debug(f'CLI: Parsed provided file into state machine:\n {fsm}')

        if NAME not in fsm.keys() or fsm[NAME].strip() == "":
            fsm[NAME] = 'StateMachine'

        code = process(fsm, language, opts[SKIP_VALIDATION])
        logging.debug(f'CLI: Processed state machine to code:\n{code}')

        filename = append_language_association(opts[FILENAME], language)

        if os.path.isabs(filename):
            out_path = filename

        else:
            out_path = os.path.join(opts[OUT_DIR], filename)

        try:
            logging.debug(f'CLI: Trying to save code in file: {out_path} and replace existing flag: {opts[REPLACE]}')
            writer.save_text_file(code, out_path, opts[REPLACE])
            logging.debug(f'CLI: Saved code to file')

            info(f'Saved generated code in: {out_path}')

        except ExistingFileError:
            warn(f'File: {out_path} already exists! Code could not be saved!')
            warn('Provide "-re" option to automatically overwrite existing files')

        info(f'Parsing of {file.name} finished without errors!')

    except ParsingError as pe:
        error(f'Content could not be parsed into a valid format!\n{pe.cause}')
    except ValidationError as ve:
        error(f'Validation of state machine logic failed: {ve.cause}')
    except UnsupportedLanguageError:
        error(f'Provided language: "{language}" is not supported!')
    except InternalException:
        error('Execution aborted!')



from meta import *
from string import Template
import logging


class CodeGenerator:
    """
    Class handling the build process of code based on a fine state machine representation.

    Requires a language template on initialization to be provided to the constructor.
    For template definition and examples see: resources/templates/template

    Constructed code follows a predefined schema:

            1. A variable/collection/class/enum holding all possible states

            2. A variable holding the current state

            3. Methods that represent transition rules upon certain events
    """

    def __init__(self, template: dict):
        self.template: dict = template
        self.config: dict = template[CONFIG]
        self.states: dict = template[STATES]
        self.state: dict = template[STATE]
        self.transitions: dict = template[TRANSITIONS]
        self.header: dict = template[HEADER]
        self.footer: dict = template[FOOTER]

    def build(self, fsm: dict) -> str:
        """
        Build code from a parsed fsm representation.

        Method will assume that the given representation is correct and perform
        no further validation.

        :param fsm: Parsed dictionary representation of a fine state machine
        :return: code corresponding to the fsm
        """

        all_nodes: list = sorted(fsm[NODES], key=lambda node: node[ID])

        states_code = self.write_all_states(all_nodes)
        logging.debug(f'GENERATOR: Generated code for all states:\n{states_code}')

        first_state: str = next(iter(all_nodes))[ID]
        state_code = self.write_initial_state(all_nodes, first_state)
        logging.debug(f'GENERATOR: Generated states variable:\n{state_code}')

        transitions: dict = group_transitions(fsm[TRANSITIONS])
        functions_code = self.write_all_functions(transitions)
        logging.debug(f'GENERATOR: Generated transition code:\n{functions_code}')

        code_blocks: list = [
            states_code,
            state_code,
            functions_code
        ]

        if self.config[USE_HEADER]:
            header = self.write_header(self.header[CONTENT], fsm[NAME])
            code_blocks.insert(0, header)

            footer = self.write_footer(self.footer[CONTENT])
            code_blocks.append(footer)

        return self.config[GAP].join(code_blocks)

    def write_header(self, header_template: Template, name: str) -> str:
        return self.substitute(header_template, name=name)

    def write_footer(self, footer_template: Template) -> str:
        return self.substitute(footer_template)

    def write_all_states(self, nodes: list) -> str:
        """
        Build code that represents all states contained in the fsm.

        All found nodes will be represented in a distinct state. All states will be added to one
        global variable (class/enum/var or equivalent).

        :param nodes: List of all nodes
        :return: Code corresponding to a state collection
        """

        item_template: Template = self.states[ITEM]
        body_template: Template = self.states[BODY]

        items: str = ''
        for index, node in enumerate(nodes):
            name: str = node[ID]
            item: str = self.substitute(item_template, name=name, value=f'{index}')

            items = ''.join([items, item])

        return self.substitute(body_template, items=items)

    def write_initial_state(self, nodes: list, backup_initial: str) -> str:
        """
        Write code that assigns an initial state to a variable.

        The variable from this code snippet will be used throughout to represent the current state.
        Provide a backup value for the initial state that will be used if no node declaring itself as initial state
        can be found.

        :param nodes: List of all nodes
        :param backup_initial: Initial state value to use if no initial node is detected
        :return: Code that handles the initial state assignment
        """

        body_template: Template = self.state[BODY]

        initial_state: str = backup_initial
        for node in nodes:
            if START in node.keys() and node[START]:
                initial_state = node[ID]
                break

        return self.substitute(body_template, value=initial_state)

    def write_all_functions(self, transitions: dict) -> str:
        """
        Write the event handling code responsible for transition between states.

        Methods needs provision of all transitions grouped by their label.

        :param transitions: Dictionary of all transitions grouped by their label
        :return: Code handling transitions between states
        """

        functions_code: list = []
        for name, assignments in sorted(transitions.items(), key=lambda item: item[0]):

            sorted_assignments = sorted(assignments, key=lambda assignment: next(iter(assignment)))

            functions_code.append(self.write_function(name, sorted_assignments))

        return self.config[GAP].join(functions_code)

    def write_function(self, name: str, assignments: list):
        """
        Write code that represents all transitions that can happen for one distinct event.

        :param name: Name of the event / transition
        :param assignments: Dictionary of all transitions for the event
        :return: Code of one method/function as transition rules
        """

        body_template: Template = self.transitions[BODY]
        item_template: Template = self.transitions[ITEM]

        items = ''

        # Handle a differing first assignment as in 'if ... elif'
        if len(assignments) > 0 and ITEM_FIRST in self.transitions.keys():
            first_item_template: Template = self.transitions[ITEM_FIRST]
            first_assignment: dict = assignments.pop(0)

            for _from, _to in first_assignment.items():
                items = self.substitute(first_item_template, name=_from, value=_to)

        for assignment in assignments:
            for _from, _to in assignment.items():
                item = self.substitute(item_template, name=_from, value=_to)
                items = ''.join([items, item])

        return self.substitute(body_template, name=name, items=items)

    def substitute(self, template: Template, name: str = None, value: str = None, items: str = None) -> str:
        """
        Internal method handling Template.safe_substitute while always providing three fixed variables:

                $indent, $state, $states placeholders of all provided TEMPLATES will be substituted.

        In addition provide name / value / items as non fixed parameters.

        :param template: String template to substitute values in
        :param name: Value for '$name' placeholder in template
        :param value: Value for '$value' placeholder in template
        :param items: Value for '$items' placeholder in template
        :return: Template substitution
        """
        indent = self.config[INDENT]
        state = self.state[NAME]
        states = self.states[NAME]

        return template.safe_substitute(indent=indent, state=state, states=states, name=name, value=value, items=items)


def group_transitions(transitions: list, group_by: str = LABEL) -> dict:
    """
    Group transitions by their label.

    Transitions need to at have the following keys:
        'label': descriptive id of transition
        'from': Start Node
        'to': End Node

    :param transitions: All transitions
    :param group_by: The key to group by, defaults to 'label'
    :return: Transitions grouped by key
    """

    grouped_transitions = dict()

    for transition in transitions:
        label = transition[group_by]
        trans = {transition[FROM]: transition[TO]}

        if label not in grouped_transitions.keys():
            grouped_transitions[label] = [trans]
        else:
            grouped_transitions[label].append(trans)

    logging.debug(f'GENERATOR: Grouped transitions to:\n{grouped_transitions}')
    return grouped_transitions



from meta import TEMPLATES, supported_languages, file_associations, code_templates, template_keys
from modules.errors import UnsupportedLanguageError
from modules import reader
from string import Template
import logging


def get_template(language: str) -> dict:
    """
    Get the dictionary containing string.Template's for the provided language.

    :param language: string for programming language
    :raises UnsupportedLanguageError if language is not supported
    :return: Dictionary with Templates
    """

    if is_unsupported_language(language):
        raise UnsupportedLanguageError

    template: dict = reader.load_resource(TEMPLATES, code_templates[language])

    for value in template.values():
        for key, string in value.items():
            if key in template_keys:
                value[key] = Template(string)

    return template


def append_language_association(file_name: str, language: str) -> str:
    """
    Append the file_type for the provided language if necessary.

    :param file_name: Filename to append association on.
    :param language: string for programming language
    :raises UnsupportedLanguageError if language is not supported
    :return: File_type for provided language
    """

    if is_unsupported_language(language):
        raise UnsupportedLanguageError

    file_association = file_associations[language]

    if not file_name.endswith(file_association):
        file_name = file_name + file_association

    return file_name


def is_unsupported_language(language: str) -> bool:
    """
    Verify that the given language is supported.
    """

    not_supported = language not in supported_languages

    if not_supported:
        logging.debug(f'TEMPLATES: Identified {language} as not supported')

    return not_supported



from meta import *
from modules.logic import process
from modules.errors import ParsingError, ValidationError, UnsupportedLanguageError
from modules import defaults, reader
from typing import IO
from distutils.util import strtobool
import logging
import eel
import json

browser_options: dict = {
    'window_size': (1200, 800),
    'extensions': ['.js', '.html'],
}

file_to_load: dict or None = None


def start(file: IO, opts: dict) -> None:
    """
    Start the local server for the Graphical User Interface.

    Method utilizing eel library to provide entry point for a GUI.
    If given FILE is None the GUI will start empty.

    :param file: File to open on start or None
    :param opts: Dictionary with available gui options
    """

    startup_options = {
        PORT: opts[PORT]
    }

    if file:
        global file_to_load
        file_to_load = reader.parse_file(file)

    eel.init(GUI_PATH, allowed_extensions=browser_options['extensions'])

    if opts[NO_WINDOW]:
        startup_options[MODE] = None

    try:
        eel.start(HTML, options=startup_options, size=browser_options['window_size'])

    except Exception:
        startup_options[MODE] = 'edge'
        eel.start(HTML, options=startup_options, size=browser_options['window_size'])


"""     GUI API

    API for the graphical user interface provided with web technologies.

    All functions listed below that are annotated with @eel.expose can be called on javascript side
    via the exposed 'eel' object.

"""


@eel.expose
def parse(graph: str, language: str, opts: str) -> (str, str):
    """
    Parse the given json string in the provided language and generate code.

    :param graph: fsm as json
    :param language: Programming language to use for code generation
    :param opts: JSON string of available options
    """

    logging.debug(f'GUI: parse called with JSON: {graph} and language: {language}')
    logging.debug(f'GUI: provided options: {opts}')

    code: str or None = None
    error: str or None = None

    try:
        fsm: dict = json.loads(graph)
        logging.debug(f'GUI: Parsed provided GRAPH into state machine:\n {fsm}')

        options: dict = json.loads(opts)
        logging.debug(f'GUI: Parsed provided OPTS into options:\n {options}')

        skip_validation = options[SKIP_VALIDATION]
        if type(skip_validation) is str:
            skip_validation = bool(strtobool(skip_validation))

        if skip_validation:
            logging.debug('GUI: Validation of internal logic will be skipped!')

        code = process(fsm, language, skip_validation)
        logging.debug(f'GUI: Processed state machine to code:\n{code}')

    except ParsingError as pe:
        error = f'Content could not be parsed into a valid format!\n{pe.cause}'
    except ValidationError as ve:
        error = f'Validation of state machine logic failed: {ve.cause}'
    except UnsupportedLanguageError:
        error = f'Provided language: "{language}" is not supported!'

    if error is not None:
        logging.error(f'GUI: {error}')
        logging.warning('GUI: Code generation aborted. Returning error!')

    return code, error


@eel.expose
def load_file() -> str:
    """
    Send the file to load to the GUI.
    """

    return json.dumps(file_to_load)


@eel.expose
def set_option(key: str, value: str) -> None:
    """
    Set options via GUI.
    """

    if defaults.set_default(key, value):
        defaults.save()


@eel.expose
def unset_option(key: str) -> None:
    """
    Unset options via GUI.
    """

    if defaults.unset_default(key):
        defaults.save()


@eel.expose
def unset_all_options() -> None:
    """
    Unset all options via GUI.
    """
    
    return defaults.unset_all()


@eel.expose
def available_languages() -> str:
    """
    Provide all available languages to the GUI.
    """

    return json.dumps({'languages': [lang for lang in supported_languages]})


@eel.expose
def available_defaults() -> str:
    """
    Provide all available defaults to the GUI.

    Defaults follow the schema below:

        [key: string]: {
            'typevalue': any,
            'current': any,
            'description': string
        }

    """
    current: dict = defaults.current_defaults
    descriptions: dict = defaults.keys_help

    processed: dict = {}

    for key, value in defaults.keys_values.items():
        if value is str:
            value = ''
        elif value is bool:
            value = False
        elif value is int:
            value = 0

        processed[key] = {
            'typevalue': value,
            'current': current[key] if key in current else None,
            'description': descriptions[key]
        }

    try:
        return json.dumps(processed)

    except TypeError as te:
        logging.warning(f'GUI: json error: {te}')


@eel.expose
def get_version() -> str:
    """
    Provide version information to the GUI.
    """

    return VERSION


@eel.expose
def get_info() -> str:
    """
    Provide tool documentation the GUI.
    """

    resource: dict = reader.load_resource('info', 'README.md')

    return json.dumps(resource[CONTENT])



from os import path, listdir

# General
VERSION: str = '0.1.2'
TOOL_NAME: str = 'Beads'


# De-/Encoding
STANDARD_ENCODING: str = 'UTF8'


# Paths
ROOT_DIR: str = path.dirname(path.abspath(__file__))
RESOURCES_PATH: str = path.join(ROOT_DIR, 'resources')
GUI_PATH: str = path.join(RESOURCES_PATH, 'ui')
HTML: str = 'index.html'

# Resource sub directories
TEMPLATES: str = 'templates'
OPTION_SUB_DIR: str = 'options'

# Default-Options
DEFAULT_OPTIONS: str = 'default_options.json'

# CLI
NO_WINDOW: str = 'no-window'
PORT: str = 'port'
SKIP_VALIDATION: str = 'skip-validation'
VERBOSE: str = 'verbose'
REPLACE: str = 'replace-existing'
OUT_DIR: str = 'out-dir'
FILENAME: str = 'filename'

help_no_window: str = 'Run the gui without opening a browser window.'
help_file_gui: str = 'Load provided FILE after setup.'
help_port: str = 'Specify the PORT to run the gui on.'
help_languages: str = 'Programming language to use for code generation.'
help_no_validation: str = 'Disable validation of provided fine state machine.'
help_verbose: str = 'Run command in verbose mode printing all debugging information.'
help_out_dir: str = 'Specify DIRECTORY to write output files into.'
help_compile_name: str = 'Specify the NAME of the output file without ending.'
help_replace_existing: str = 'Automatically replace existing files with same name.'
help_set_option: str = 'Set a default runtime option: OPTION VALUE'
help_unset_option: str = 'Delete a default runtime option: OPTION'
help_unset_all: str = 'Delete all default runtime options.'
help_show_options: str = 'Show all available default options.'


# FSMs
NODES: str = 'nodes'
START: str = 'start'
TRANSITIONS: str = 'transitions'
ID = 'id'
VALUE = 'value'
LABEL: str = 'label'
FROM: str = 'from'
TO: str = 'to'


# Templates
CONFIG: str = 'config'
USE_HEADER: str = "useHeader"
GAP: str = 'gap'
INDENT: str = 'indent'
NAME: str = 'name'
BODY: str = 'body'
ITEM: str = 'item'
ITEM_FIRST: str = 'item_first'
STATES: str = 'states'
STATE: str = 'state'
HEADER: str = 'header'
FOOTER: str = 'footer'

template_keys: set = {BODY, ITEM, ITEM_FIRST, "content"}
TEMPLATE_FILE: str = '.template'
LANGUAGE: str = 'language'
FILE_TYPE: str = 'file_type'
TEMPLATE_VERSION: str = 'version'
TEMPLATE_VERSIONS: set = {'1'}

CONTENT: str = 'content'
MODE: str = 'mode'

# Programming languages
supported_languages: set = set()
code_templates: dict = dict()
file_associations: dict = dict()


def __find_languages__() -> None:
    """
    Find and set all supported languages via the existence of a valid template
    """
    from modules.reader import read_config

    global supported_languages
    global code_templates
    global file_associations

    for file_path in __gather_template_files__():

        with open(file_path) as content:

            first_line = content.readline()
            if first_line.startswith('#!'):

                config: dict = read_config(first_line)

                if __is_valid_config__(config):
                    language = config[LANGUAGE]
                    supported_languages.add(language)
                    code_templates.update({language: ''.join([language, TEMPLATE_FILE])})
                    file_associations.update({language: config[FILE_TYPE]})


def __is_valid_config__(config: dict) -> bool:
    """
    Verify that all needed keys are present and the declared version is supported

    :param config: Dictionary with configuration key:values
    :return: True if verified, otherwise False
    """

    required_keys = [TEMPLATE_VERSION, LANGUAGE, TEMPLATE_FILE, FILE_TYPE]

    return [key in config.keys() for key in required_keys] and config[TEMPLATE_VERSION] in TEMPLATE_VERSIONS


def __gather_template_files__() -> list:
    """
    Generator for file_paths of template files.

    :return: File_paths that are files and end with '.template'
    """
    templates_path = path.join(RESOURCES_PATH, TEMPLATES)

    for file in listdir(templates_path):

        file_path: str = path.join(templates_path, file)

        if path.isfile(file_path) and file.endswith(TEMPLATE_FILE):
            yield file_path


__find_languages__()



#!/usr/bin/env python
import unittest
import beads.modules.reader as reader
from modules.errors import ParsingError
from io import BytesIO, TextIOWrapper


class BeadsParsingTest(unittest.TestCase):
    """
    Test class to test correct handling of .bead files
    """

    def test_correct_file_content(self):
        """
        Test that a correct file content can be parsed without error
        """
        self.maxDiff = None

        content_lines: bytes = \
            b"A:Transition_1:B\n"\
            b"A:Transition_2:C\n"\
            b"B:Transition_1:D\n"\
            b"B:Transition_3:A\n"\
            b"C:Transition_4:B"

        wrapper: TextIOWrapper = TextIOWrapper(BytesIO(content_lines))

        expected: dict = {
            'name': 'graph',
            'nodes': [
                {'id': 'A'},
                {'id': 'B'},
                {'id': 'C'},
                {'id': 'D'}
            ],
            'transitions': [
                {'label': 'Transition_1', 'from': 'A', 'to': 'B'},
                {'label': 'Transition_2', 'from': 'A', 'to': 'C'},
                {'label': 'Transition_1', 'from': 'B', 'to': 'D'},
                {'label': 'Transition_3', 'from': 'B', 'to': 'A'},
                {'label': 'Transition_4', 'from': 'C', 'to': 'B'}
            ]
        }

        actual: dict = reader.parse_beads(wrapper)

        self.assertEqual(actual, expected)

    def test_config_can_be_parsed(self):
        """
        Test that a config can be parsed without error
        """
        self.maxDiff = None

        content_lines: bytes = \
            b"#! name:GRAPH start:C\n"\
            b"A:Transition_1:B\n"\
            b"A:Transition_2:C\n"\
            b"B:Transition_1:D\n"\
            b"B:Transition_3:A\n"\
            b"C:Transition_4:B"

        wrapper: TextIOWrapper = TextIOWrapper(BytesIO(content_lines))

        expected: dict = {
            'name': 'GRAPH',
            'nodes': [
                {'id': 'A'},
                {'id': 'B'},
                {'id': 'C', "start": True},
                {'id': 'D'}
            ],
            'transitions': [
                {'label': 'Transition_1', 'from': 'A', 'to': 'B'},
                {'label': 'Transition_2', 'from': 'A', 'to': 'C'},
                {'label': 'Transition_1', 'from': 'B', 'to': 'D'},
                {'label': 'Transition_3', 'from': 'B', 'to': 'A'},
                {'label': 'Transition_4', 'from': 'C', 'to': 'B'}
            ]
        }

        actual: dict = reader.parse_beads(wrapper)

        self.assertEqual(actual, expected)

    def test_config_ignores_unsupported_values(self):
        """
        Test that a config ignores unsupported values
        """
        self.maxDiff = None

        content_lines: bytes = \
            b"#! title:GRAPH beginning:C\n"\
            b"A:Transition_1:B\n"\
            b"A:Transition_2:C\n"\
            b"B:Transition_1:D\n"\
            b"B:Transition_3:A\n"\
            b"C:Transition_4:B"

        wrapper: TextIOWrapper = TextIOWrapper(BytesIO(content_lines))

        expected: dict = {
            'name': 'graph',
            'nodes': [
                {'id': 'A'},
                {'id': 'B'},
                {'id': 'C'},
                {'id': 'D'}
            ],
            'transitions': [
                {'label': 'Transition_1', 'from': 'A', 'to': 'B'},
                {'label': 'Transition_2', 'from': 'A', 'to': 'C'},
                {'label': 'Transition_1', 'from': 'B', 'to': 'D'},
                {'label': 'Transition_3', 'from': 'B', 'to': 'A'},
                {'label': 'Transition_4', 'from': 'C', 'to': 'B'}
            ]
        }

        actual: dict = reader.parse_beads(wrapper)

        self.assertEqual(actual, expected)

    def test_parsing_handles_whitespaces_correct(self):
        """
        Test that parsing of .bead files handles whitespaces correctly
        """
        self.maxDiff = None

        content_lines: bytes = \
            b"A :Transition_1 :B\n"\
            b"A: Transition_2: C\n"\
            b"B : Transition_1 : D\n"\
            b"B   : Transition_3 :    A\n"\
            b"C: Transition_4 :B"

        wrapper: TextIOWrapper = TextIOWrapper(BytesIO(content_lines))

        expected: dict = {
            'name': 'graph',
            'nodes': [
                {'id': 'A'},
                {'id': 'B'},
                {'id': 'C'},
                {'id': 'D'}
            ],
            'transitions': [
                {'label': 'Transition_1', 'from': 'A', 'to': 'B'},
                {'label': 'Transition_2', 'from': 'A', 'to': 'C'},
                {'label': 'Transition_1', 'from': 'B', 'to': 'D'},
                {'label': 'Transition_3', 'from': 'B', 'to': 'A'},
                {'label': 'Transition_4', 'from': 'C', 'to': 'B'}
            ]
        }

        actual: dict = reader.parse_beads(wrapper)

        self.assertEqual(actual, expected)

    def test_parsing_does_not_allow_nonalpha_characters(self):
        """
        Test that parsing .bead files does not allow non alpha characters other than _
        """
        self.maxDiff = None

        content_lines: bytes = \
            b"A:Transition_1:B\n"\
            b"A:Transition_2:!C\n"\

        wrapper: TextIOWrapper = TextIOWrapper(BytesIO(content_lines))

        self.assertRaises(ParsingError, lambda: reader.parse_beads(wrapper))



class InternalException(BaseException):
    """ Error Base class providing exception management on highest level """
    pass


class ValidationError(InternalException):
    """ Any error that occurs while validating the internal logic of a fine state machine """
    def __init__(self, cause):
        self.cause = cause


class UnsupportedLanguageError(InternalException):
    """ Error indicating that the provided language option is not supported """


class ParsingError(InternalException):
    """ Any error that occurs while trying to parse the provided input """
    def __init__(self, cause):
        self.cause = cause


class ExistingFileError(InternalException):
    """ Error indicating that a file with the same name already exists and the '-re' flag is not set """



from setuptools import find_packages, setup
import pathlib

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='Beads',
    version='0.1.2',
    author='Jakob Baatz, Rico Possienka, Pavel Nepke, Marco Wenning, Adrian Wuillemet',
    description='A commandline tool to parse fine state machines into code.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://gitlab.beuth-hochschule.de/s40242/stategen',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['click', 'eel', 'markdown'],
    entry_points={
        'console_scripts': [
            'beads = beads.main:beads'
        ]
    },
    keywords="state machine code generation gui",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from modules.errors import ExistingFileError
from meta import RESOURCES_PATH
import logging
import json
import os


def save_text_file(text: str, path: str, overwrite_existing: bool = False) -> None:
    """
    Write the provided TEXT into a file with given PATH.

    Will only overwrite already existing files if OVERWRITE_EXISTING is True.

    :param text: Content to save
    :param path: File path to save into
    :param overwrite_existing: Flag to specify overwriting of existing files
    :raises ExistingFileError if file with PATH exists but OVERWRITE_EXISTING is False
    :raises IOError
    """

    if overwrite_existing:
        overwrite(text, path)

    else:
        try:
            open(path)
            logging.debug(f'WRITER: file with path {path} could be opened. File exists')
            raise ExistingFileError

        except IOError:
            overwrite(text, path)


def save_resource(subdir: str, name: str, full_content: dict) -> None:
    """
    Write the provided FULL_CONTENT into a file with given NAME.

    :param subdir: Sub directory of the resources folder
    :param name: Filename
    :param full_content: Full resources content as dict
    """

    resource_path: str = os.path.join(RESOURCES_PATH, subdir, name)

    content: str = json.dumps(full_content)

    save_text_file(content, resource_path, True)


def overwrite(text: str, path: str) -> None:
    """
    Write TEXT into PATH in 'w+' mode.
    Will overwrite existing files!

    :param text: Content to write/save
    :param path: PATH to save into
    :raises IOError
    """

    out_dir: str = os.path.dirname(path)

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    output_file = open(path, 'w+')
    output_file.write(text)
    output_file.close()

    logging.debug(f'WRITER: Writing successful for file: {path}')




from meta import *
from modules.errors import ValidationError
from modules.generator import group_transitions
import logging


def inspect(fsm: dict) -> None:
    """
    Inspect the provided fsm and assert the correctness of its internal logic.

    Provide a dictionary that represents a fine state machine as:

        {
            name: "",
            nodes: [
                {id: , ...}
            ],
            transitions: [
                {label: , from: , to: , ...}
            ]
        }

    All above displayed keys are required to pass validation. Other keys are optional and allowed.

    :param fsm: Fine state machine parsed as dictionary.
    :return: True if every assertion is true
    :raises ValidationError on any invalid logic
    """
    try:
        # assert fsm contains necessary keys
        keys = fsm.keys()

        assert NAME in keys, f'Missing required key: {NAME}'
        assert NODES in keys, f'Missing required key: {NODES}!'
        assert TRANSITIONS in keys, f'Missing required key: {TRANSITIONS}!'

        # assert fsm contains at least one node
        nodes: list = fsm[NODES]
        assert len(nodes) > 0, 'No state is defined!'

        # assert fsm contains at least one transition
        transitions: list = fsm[TRANSITIONS]
        assert len(transitions) > 0, 'No transition is defined!'

        # assert all nodes have the required 'id' key
        faulty_nodes = nodes_with_no_id(nodes)
        assert len(faulty_nodes) == 0, f'{len(faulty_nodes)} node(s) are missing the required key "{ID}"'

        # assert id of nodes is unique
        duplicate_nodes: list = duplicated_nodes(nodes)
        assert len(duplicate_nodes) == 0, f'These node(s) have been declared more than once: {duplicate_nodes}'

        # assert all transitions have the required keys 'label', 'from', 'to'
        faulty_transitions: list = transitions_with_missing_keys(transitions)
        assert len(faulty_transitions) == 0, f'Required keys [{LABEL},{FROM},{TO}] missing' \
            f' in transitions: {faulty_transitions}'

        # assert maximum of one start node
        start_nodes, start_node = identify_start_nodes(nodes)
        assert len(start_nodes) < 2, f'More than one start state: {start_nodes}'

        transitions_grouped_by_from: dict = group_transitions(transitions, group_by=FROM)
        start_node = start_node if start_node is not None else next(iter(nodes))[ID]

        # assert that all nodes are reachable
        isolated_nodes: set = unreachable_nodes(nodes, transitions_grouped_by_from, start_node)
        assert len(isolated_nodes) == 0, f'Not all nodes can be reached: {isolated_nodes}'

        transitions_grouped_by_label: dict = group_transitions(transitions, group_by=LABEL)

        # assert that for each transition label a from node exists only once
        conflicting_transition_list: list = conflicting_transitions(transitions_grouped_by_label)
        assert len(conflicting_transition_list) == 0, f'Conflicting transitions for: {conflicting_transition_list}'

        # assert all nodes referenced in transitions are declared
        undeclared_nodes: set = inspect_transitions_for_undeclared_nodes(nodes, transitions)
        assert len(undeclared_nodes) == 0, f'States referenced but not declared: {undeclared_nodes}'

    except AssertionError as ae:
        logging.debug('VALIDATION: Failed due to AssertionError')
        raise ValidationError(ae)


def unreachable_nodes(nodes: list, grouped_transitions: dict, start_node: str) -> set:
    """
    Transition over the NODES starting with the initial START_NODE via the GROUPED_TRANISIONS and return any node that
    was not visited in the process.

    :param nodes: List of all nodes
    :param grouped_transitions: Dictionary of all transitions grouped by their FROM node
    :param start_node: Node to start from
    :return: List of nodes in NODES that were not visited
    """

    isolated_nodes: set = set(node[ID] for node in nodes)
    isolated_nodes.discard(start_node)

    node_queue: list = [start_node]

    while len(node_queue) > 0:
        node: str = node_queue.pop(0)

        if node in grouped_transitions.keys():
            associated_transitions: list = grouped_transitions[node]

            for to_node in associated_transitions:
                if isolated_nodes.__contains__(to_node[node]):
                    isolated_nodes.discard(to_node[node])
                    node_queue.append(to_node[node])

    return isolated_nodes


def conflicting_transitions(transitions_grouped_by_label: dict) -> list:
    """
    Group all labels of TRANSITIONS that are declared multiple times from the same state.

    :param transitions_grouped_by_label: All Transitions grouped by their label
    :return: List of labels of conflicting transitions
    """

    conflicting_transitions_list: list = []

    for label, transitions_for_label in transitions_grouped_by_label.items():
        unique_ids = set(next(iter(t)) for t in transitions_for_label)

        if len(unique_ids) != len(transitions_for_label):
            conflicting_transitions_list.append(label)

    return conflicting_transitions_list


def identify_start_nodes(nodes: list) -> (list, str or None):
    """
    List all node ids of nodes that are declared as start nodes.

    :param nodes: All nodes
    :return: List of nodes declared as start, the last found start node or None if none was found
    """
    start_nodes: list = []
    start_node: str or None = None

    for node in nodes:
        if START in node.keys() and node[START] is True:
            start_nodes.append(node[ID])
            start_node = node[ID]

    return start_nodes, start_node


def duplicated_nodes(nodes: list) -> list:
    """
    List all nodes that have been declared more than once.

    :param nodes: All nodes
    :return: List of duplicated node ids
    """

    duplicate_nodes: list = []
    found_nodes: set = set()

    for node in nodes:
        node_id = node[ID]
        if node_id in found_nodes:
            duplicate_nodes.append(node_id)
        else:
            found_nodes.add(node_id)

    return duplicate_nodes


def nodes_with_no_id(nodes: list) -> list:
    """
    List of all nodes without ID key

    :param nodes: All nodes
    :return: List of nodes without ID key
    """
    faulty_nodes: list = []

    for node in nodes:
        if ID not in node.keys():
            faulty_nodes.append(node)

    return faulty_nodes


def transitions_with_missing_keys(transitions: list) -> list:
    """
    List transitions that miss at least one of the required keys:
        'label', 'from', 'to'

    :param transitions: All transitions
    :return: List of transitions missing a required key
    """

    faulty_transitions: list = []

    for transition in transitions:
        if {LABEL, FROM, TO} > transition.keys():
            faulty_transitions.append(transition)

    return faulty_transitions


def inspect_transitions_for_undeclared_nodes(nodes: list, transitions: list) -> set:
    """
    List all nodes that have been referenced in a transition but haven't been declared.

    :param nodes: All nodes
    :param transitions: All transitions
    :return: List of node ids that are referenced but not declared
    """
    declared_nodes: set = set(node[ID] for node in nodes)
    undeclared_nodes: set = set()

    for transition in transitions:
        _from = transition[FROM]
        _to = transition[TO]

        if _from not in declared_nodes:
            undeclared_nodes.add(_from)

        if _to not in declared_nodes:
            undeclared_nodes.add(_to)

    return undeclared_nodes
